# required by troff2html
# John M. Troyer
# Copyright 1994 The Regents of the University of California
# please see the file named COPYRIGHT

# changes troff non-ascii characters and special symbols into 
# iso8859 html entities
# (see http://www.uni-passau.de/~ramsch/iso8859-1.html)

# right now, delete printing characters we can't convert'
# might want to make this command-line settable -- delete or leave alone


# this must be called early, so when we start slinging html commands around,
#  they don't get escaped, too.
sub htmlescape {
    # escape & < and >:
    s/\&/\&amp;/g;
    s/</\&lt;/g;
    s/>/\&gt;/g;
}
# I forget exactly what breaks when these are moved to inline_substitutions,
# with the rest of the predefined strings, but something does.
sub footnotemarks {
    if ($command =~ /^\([fdx]$/) {
	# a reverse link back to text
	$text =~ s/\\\*\*/&mark_in_note("f")/ge;	
	$text =~ s/\\\*#/&mark_in_note("d")/ge;
    } else {
	$currentfoot++ if ($command eq ')f');

	$text =~ s/\\\*\*/&mark_outof_note("f")/ge;	
	$text =~ s/\\\*#/&mark_outof_note("d")/ge; # number incremented here
    }
}

# this routine calls the rest of the subroutines
sub escapechars {
    &mepredefinedstrings;
    &mespecialchars;
    &troffescapes;
    &troffnonascii;
    
    # Mosaic currently doesn't support all entities. Unless $opt_entity '
    # is set, just use Mosaic-understandable set.
    if (!$opt_entity) {
	&mosaic_bad_dog;
    }
}

# me predefined strings

    
sub mepredefinedstrings {

    # \**, \*# handled in &footnotemarks

    # superscripts, subscripts handled like nroff. [1] [2] yuck x<i>
    # any better ideas? ^ and _?
    # I'm assuming \*{, \*} are the same as \*[, \*], although I haven't
    #   seen it documented.
    # BUG: 1,2 comes out [1][,][2] instead of [1,2]
    if ($superscripts eq "nroff") {
	s/\\\*\[/\[/g; 
	s/\\\*\]/\]/g;
	s/\\\*\&lt;/&lt;/g;
	s/\\\*\&gt;/&gt;/g;
	s/\\\*\{/\[/g;
	s/\\\*\}/\]/g;
    } elsif ($superscripts eq "math") {
	s/\\\*\[/\^/g; 
	s/\\\*\]//g;
	s/\\\*\&lt;/\_/g;
	s/\\\*\&gt;//g;
	s/\\\*\{/\^/g;
	s/\\\*\}//g;
    }
    # dates
    s/\\\*\(dw/$weekday/g;
    s/\\\*\(mo/$month/g;
    s/\\\*\(td/$month $mday, $year/g;
    s/\\n\(dy/$mday/g;
    s/\\n\(mo/$mon/g;
    s/\\n\(yr/$year/g;
    # quotes
    s/\\\*\(lq/\"/g; #"
    s/\\\*\(rq/\"/g; #"
    s/\\\*-/--/g;
}

sub mespecialchars {
    # these translate to appropriate entities &Aacute;, &aacute;, etc.
    # obviously they'll break if somebody tries to do a &Zacute; or other'
    #  non-defined entity.
    # we include special cases in case somebody is just using the accents
    #  on no letters
    s/(.)\\\*'/&$1acute;/g; #'
    s/\\\*'/'/g; 
    s/(.)\\\*`/&$1grave;/g; #`
    s/\\\*`/`/g; #`
    s/(.)\\\*:/&$1uml;/g; 
    s/(.)\\\*~/&$1tilde;/g;
    s/\\\*~/~/g;
    s/(.)\\\*,/&$1cedil;/g;
    s/(.)\\\*o/&$1ring;/g;
    s/\\\*o/&deg;/g;
    s/(.)\\\*^/&$1circ;/g;

    # these are not defined. Just drop the accent.
    warn("Character $1 czech not iso8859; can't print it\n") if s/(.)\\\*v/\1/;

    # delete these
    warn("Character 'There exists' not iso8859; can't print it\n") 
	if s/\\\*\(qe//;
    warn("Character 'For all' not iso8859; can't print it\n") 
	if s/\\\*\(qa//;
}

sub troffescapes {
    # continue 
    s/\\c\b//g;

    # escape sequences; 
    s/\\'/'/g; 
    s/\\`/`/g;
    s/\\-/-/g;
    s/\\\././g;
    # should be hard space
    s/\\ /&nbsp;/g;
    s/\\0/ /g;
    s/\\\|//g; # following NROFF convention
    s/\\\^//g; # following NROFF convention
    s/\[,\]//g; # [1][,][2] to [1][2] better than nuthin
    #s/\\\&//g;
    s/\\\&amp;//g; #note this one has already been escaped to html!
    # eventually, size changes could be used as heuristics for headers,
    # but for now, delete them.
    s/\\s[+\-]?\d+//g;
    s/\\s//g;

    # line drawing
    s/\\l\'\S+/\<hr\>/;
    # most of the troff sequences are NOT supported and are silently killed here
    # delete register references; will play hell with macro definitions
    s/(\\n\(..)//g;
    # delete all other escape sequences; we don't know what to do with them
    s/(\\\*\(..)//g;
}

sub troffnonascii {
    s/\\\(em/--/g;
    s/\\\(hy/-/g;
    # if bullets are used in lists, we can catch them. These are in text
    # replace them with middot for now...poor substitute
    s/\\\(bu/&middot;/g;
    warn("square not iso8859; can't print it\n") if s/\\\(sq//g;
    s/\\\(ru/_/g;
    s/\\\(14/&frac14;/g;
    s/\\\(14/&frac12;/g;
    s/\\\(14/&frac34;/g;

    # ligatures
    s/\\\(fi/fi/g;
    s/\\\(fl/fl/g;
    s/\\\(Fi/ffi/g;
    s/\\\(Fl/ffl/g;
    s/\\\(de/&deg;/g;
    # substitute asterisk for dagger, since it's often used it footnotes
    #  we can't just delete it
    warn("converting dagger to asterisk; sorry\n") if s/\\\(dg/*/g;
    s/\\\(fm/'/g; #'
    s/\\\(ct/&cent;/g;
    s/\\\(rg/&copy;/g;
    s/\\\(co/&reg;/g;

    # special font
    s/\\\(pl/+/g;
    s/\\\(mi/-/g;
    s/\\\(eq/=/g;
    s/\\\(\*\*/*/g;
    s/\\\(sc/&sect;/g;
    s/\\\(aa/'/g; #'
    s/\\\(ga/`/g; #`
    s/\\\(ul/_/g;
    s/\\\(sl/\\/g;
    
    # greek -- convert to english letters for now
    # should be settable option -- could switch to name!
    # we have mu, and we'll fake sz for beta '
    s/\\\(\*b/&szlig;/g;
    s/\\\(\*m/&micro;/g;
    s/\\\(\*(\w)/\1/g;

    # math many liberties taken here
    warn("warning: changing sqrt sign to 'sqrt'\n") if s/\\\(sr/sqrt/g;
    s/\\\(rn/&hibar;/g; # root sign extender
    s/\\\(&gt;=/&gt;=/g;
    s/\\\(&lt;=/&lt;=/g;
    s/\\\(==/==/g;
    s/\\\(~=/~=/g;
    s/\\\(ap/~/g;
    s/\\\(!=/!=/g;
    s/\\\(-&gt;/-&gt;/g;
    s/\\\(&lt;-/&lt-/g;
    s/\\\(ua/^/g;
    s/\\\(da/v/g;
    s/\\\(mu/&times;/g;
    s/\\\(di/&divide;/g;
    s/\\\(\+\-/&plusmn;/g;
    s/\\\(no/&not;/g;
    s/\\\(or/\|/g;
    s/\\\(ci/o/g;
    # big brackets we'll just make into lots little ones... '
    s/\\\(lt/\{/g;
    s/\\\(lb/\{/g;
    s/\\\(rt/\}/g;
    s/\\\(rb/\}/g;
    s/\\\(lk/\{/g;
    s/\\\(rk/\}/g;
    s/\\\(bv/\|/g;
    s/\\\(lf/\[/g;
    s/\\\(rf/\]/g;
    s/\\\(lc/\[/g;
    s/\\\(rc/\]/g;

    # delete characters that we can't print'.
    warn("deleting non-ascii char $1\n") if s/(\\\(\S\S)//;
}

# mosaic doesn't understand some entities (taken from XMosaic 2.4)

sub mosaic_bad_dog {
    s/&nbsp;/&#160;/g;
    s/&iexcl;/&#161;/g;
    s/&cent;/&#162;/g;
    s/&pound;/&#163;/g;
    s/&curren;/&#164;/g;
    s/&yen;/&#165;/g;
    s/&brkbar;/&#166;/g;
    s/&sect;/&#167;/g;
    s/&uml;/&#168;/g;
    s/&copy;/&#169;/g;
    s/&ordf;/&#170;/g;
    s/&laquo;/&#171;/g;
    s/&not;/&#172;/g;
    s/&shy;/&#173;/g;
    s/&reg;/&#174;/g;
    s/&hibar;/&#175;/g;
    s/&deg;/&#176;/g;
    s/&plusmn;/&#177;/g;
    s/&sup2;/&#178;/g;
    s/&sup3;/&#179;/g;
    s/&acute;/&#180;/g;
    s/&micro;/&#181;/g;
    s/&para;/&#182;/g;
    s/&middot;/&#183;/g;
    s/&cedil;/&#184;/g;
    s/&supl;/&#185;/g;
    s/&ordm;/&#186;/g;
    s/&raquo;/&#187;/g;
    s/&frac14;/&#188;/g;
    s/&frac12;/&#189;/g;
    s/&frac34;/&#190;/g;
    s/&iquest;/&#191;/g;
    s/&Dstrok;/&#208;/g;
    s/&times;/&#215;/g;
    s/&divide;/&247;/g;
}


1;

