# handlemacros.pl
# John M. Troyer
# Copyright 1994 The Regents of the University of California
# please see the file named COPYRIGHT

# do the actual HTML translation and output
# this file is required by the main troff2html

#### paragraphs

# called by .lp, .pp
# we go back to roman every time here. 
# (How is the default font changed in roff?)
# if headers use lp, you'll have to mark them up by hand. Bad dog!

# this is called by all paragraph-making macros
sub para {
    $pretext .= "</blockquote>" if ($quotedpara);
    if ($command eq 'pp' || $command eq 'lp' || $command eq 'bp') {
	$pretext .= &pop_all_lists;
    }
    if (!$intable) {
	$pretext .= &pop_all_fonts . "<p>";
    }

}

# sp
# gives only para; no font popping, etc.
sub sppara {
    $pretext .="<p>" if !$intable;
}


# QP
sub quotepara {
    $pretext = "<blockquote>";
    $quotedpara = 1;
}

sub indentpara {
    # get indented word, if there is one. Drop optional indent size
    local(@args) = @_ if @_;

    # I might want to call para here for the <p>
    $pretext .= &pop_all_fonts if !$intable;

    #sometimes people make bulleted lists with ip
    if ($args =~ s/^\\\(bu// ||
	$args =~ s/^\*([^\*])/\1/) {
	$pretext .= &listitem("bu") . "<li>$args[0]";
    } else {
	$pretext .= &listitem("ip") . "<dt>$args[0]<dd>";
    }
    $pretextprints = 1;
}

# ms macros AL AR 2P
# we're going to assume somebody wanted a <dl> list
# I have no idea how this macro is usually used
sub ALpara {
    local($command) = @_ if @_;

    &para;

    # double label paragraphs just get labels shoved together
    if ($command eq '2P') {
	@args = ("$args[0] $args[1]");
    }

    # much like ms macro, don't check if text is long enough :-)
    local(@text) = split('\n',$text);
    local($newtext);

    while (@args) {
	$pretext .= &listitem("ip") . "<dt>" . shift(@args) . "<dd>" .
	    shift(@text) . "\n";
    }
    # now rewrite global text with what's left over...
    $text = join('\n',@text);
}
    
sub list {
    local($command) = @_ if @_;

    $pretext .= &listitem($command) . "<li>";
}

# list stack code from ms2html; changed to return, not print

# check if need to push or pop a list level
# when a new list item appears:
sub listitem {
    local($ltype) = @_;     # this list item type
    local($result);

    if ($#lstack < 0) {
	# no current list, so start new list:
	$result = &newlist($ltype);
    }
    elsif ($lstack[$#lstack] ne $ltype) {
	if (($#lstack > 0) && ($lstack[$#lstack - 1] eq $ltype)) {
	    # print STDERR "Popping from $ltype to $lstack[$#lstack-1]\n";
	    $result = &poplist;
	}
	else {
	    $result = &newlist($ltype);
	}
    }
    return $result;
}

# start a new list:
sub newlist {
    local($ltype) = @_;     # this list item type
    local($result);

    &para; # hope this is right.

    if ($ltype eq "np") { $result = "<OL>"; }
    elsif ($ltype eq "bu") { $result = "<UL>"; }
    elsif ($ltype eq "ip") { $result = "<DL>"; }
    push(@lstack,$ltype);
    
    return $result;
}

# pop the current list:
sub poplist {
    local($ltype);
    local($result);

    $ltype = pop(@lstack);
    if ($ltype eq "np") { $result =  "</OL>"; }
    elsif ($ltype eq "bu") { $result = "</UL>"; }
    elsif ($ltype eq "ip") { $result = "</DL>"; }
    else { print STDERR "poplist error: unknown list type \"$ltype\"\n"; }
    # should never happen!
    return $result;
}

# pop out of all remaining lists:
sub pop_all_lists {
    local($result);

    while ($#lstack >= 0) {
	$result .= &poplist;
    }

    return $result;
}

#### sections

# sh is optional number, title, optional depths
sub section {
    local($command) = @_ if @_;

    &para; # pop all fonts, etc.
    $pretext .= &pop_all_lists;

    if ($command eq 'uh') {
	$title = join(' ',@args);
	$id = $currentunnumberedsection++;
	# don't mess with headerlevel ?
	$headerlevel = 1 if (!$headerlevel);
    } else {
	# command eq 's'

	# if the first argument is a number, that is new headerlevel
	# remember headerlevel goes from 1-6, not 0-5
	$headerlevel = shift @args if $args[0] =~/^\d+$/;

	# title is next arg
	$title = shift @args;
	
	# increment current counter
	$currentsection[$headerlevel]++;
	
	# and reset all counters underneath it to zero
	if ($headerlevel < 6) {
	    foreach $i ($headerlevel+1..6) {
		$currentsection[$i] = 0;
	    }
	}

	# if counters explicitly set, grab them
	# if counters set with a number register, don't bother
	$i = 0;
	while(@args) {
	    $i++;
	    $temp = shift @args;
	    if ($temp !~ /\D/) {
		$currentsection[$i] = $temp;
	    }
	}


	# get a label of 1st through headerlevel numbers
	$id = join(".",@currentsection[1..$headerlevel]);
    }

    # unique id sh or uh
    $name = "$command-$id";
    
    # eg: 1.2.3.4 Section title here
    $title = "$num $title" if ($command eq "sh" && $opt_numbersection);

    # handles file splitting and printing heading <hn> 
    &startnewsection($headerlevel, "$basename-$name.html", $title, $name);
}


# if +, is this relative? if so, isn't implemented, but easy to
sub setdepth {
    local($args) = @_ if @_;

    $headerlevel = $args;
}


#### displays
# called for (l (b (z 
# puts a <br> after every line unless F is given
# for L (left) M (margin) lists, just use paragraphs <p>
# floating keeps (z) default to M
# for other lists (l b), uses <dl><dd> to emulate I (default) and C both

# begindisplay("(l") can emulate a call to whatever command is given as arg

sub begindisplay {
    local($command) = @_ if @_;

    # sometimes people make bulleted lists by hand with bullets or asterisks
    # this code should be pulled out & optional, since it makes mistakes
    $* = 1;  # ^ matches every \n
    if ($text =~ /^\s*\\\(bu/ ||
	($text =~ /^\s*\*/ && $text !~ /^\s*\*\s*\*/)) {
	0 while ($text =~ s/^\s*(\\\(bu|\*)/\<li\>/);
	$pretext .= &listitem("bu");
	$* = 0;
	return;
    }
    $* = 0;

    push(@breakstack,$breaklines);
    if (grep(/F/,@args)) {
	$breaklines = 0;
    } else {
	$breaklines = 1;
    }

    if (grep(/[LM]/,@args) || $command eq '(z') {
	$pretext = "<p>";
	$endblock = "<p>";
    } else { 
	# default is I
	# can't center so C goes here too
	$pretext = "<dl><dd>";
	$endblock = "</dl>";
    }
}

# here I am popping all fonts but not all lists. I hope that's correct
sub enddisplay {
    local(@command) = @_ if @_;

    $breaklines = pop(@breakstack);
    
    $pretext = &pop_all_fonts . $endblock;

    $endblock = "";

}

# we can't center yet till HTML 3.0
# for now we just indent, making sure we don't indent twice if .(b .(c
sub begincenter {
    $pretext = &pop_all_lists;
    # push since .(c can be inside another block
    push(@breakstack,$breaklines);
    $breaklines = 1;
    if ($endblock ne "</dl>") {
	$pretext .= "<dl><dd>";
	$endblock = "</dl>";
    }
}

# here we don't pop everything. No idea if that's right
sub endcenter {
    $breaklines = pop(@breakstack);
    $pretext = $endblock;
    $endblock = "";
}

# now BSD manuals don't explictly say you can embed a (q inside
#  another block, so I'm assuming you can't
sub beginquote {
    $pretext = "<blockquote>";
    push(@breakstack,$breaklines);
    $breaklines = 0;
}

sub endquote {
    $pretext = "</blockquote>";
    $breaklines = pop(@breakstack);
}

##### annotations

# both pushdelayed and popdelayed need to be rewritten
# preproccesor properties need to be consolidated in one place

## delayed text is temporarily redirected to a temp file

# called with (x (f (d (well, actually x, f, d)
# or with preproc "eqn" "tbl"
sub pushdelayed {
    local($command, $args) = @_ if @_;

    # make a unique identifier
    # have a separate counter for EACH type of delayed text
    # some of these files (postscript tbl, eqn) will stay around,
    #  and so this gives us name-tbl.1, 2, 3, etc., not name-tbl.3, 8, 22
    $delid{$command}++;
    (local($name) = $command) =~ s/^\(//;
    $name = "$basename-$name.$delid{$command}";

    # start writing to a temp file
    # first get rid of $pretext in old file
    &output($pretext);
    $pretext = "";
    local($fh) = "FILE$name";
    open($fh,">$name") || die("can't open tempfile $name: $!\n");
    push(@fstack,select($fh));	
    push(@dstack,$name);
    push(@textonpagestack,$textonpage);

    # indices
    if ($command eq 'x') {
	$currentindex = $args ? $args : 'x';

    # delayed text (d is sometimes used for footnotes )
    } elsif ($command eq 'd') {
	$delayedtext .= "<dt>";

    # footnotes
    } elsif ($command eq 'f') {
	$reftext .= "<dt>";
    
    # tables
    } elsif (grep(/$command/,keys %preproc)) {
	push(@breakstack,0);
	$breakstack = 0;

	# this only hits for tables
	# if doublebox, sending through tbl | nroff | col -b clips the
	# first char of each line. escape this. remove for HTML+, obviously
	if ($text =~ /doublebox/ && $preproc{$command} eq 'nroff') {
	    $doubleboxhack = 1;
	}
	# intable is going to be an indicator for 
	# 1. no <p> at sp
        # 2. if doublebox, put a blank in front of each line, 
	#    so nroff | col -b does eat one
	# note that $intable doesn't refer to the current $text, since that'
	# will be tbl options and formatting commands -- don't want to '
	# escape these
	$intable = 1;

	# by the time pushdelayed gets the text, it's been html escaped.
	# that's a no-no for sections destined for postscript
	$text = &unescape($text) if ($preproc{$command} eq 'ps' && 
				     $args ne 'inline');

    } else {
	warn("command $command ignored\n");
    }
}

# this is getting messy, with 3 exits in the postscript, it too long, 
# and could probably stand to be rewritten :-)

sub popdelayed {
    local($command, $args) = @_ if @_;

    # restore old default filehandle
    local($fh) = select(pop(@fstack));
    
    local($thisfile) = pop(@dstack);
    close($fh) || die("can't close tempfile $thisfile");

    $textonpage = pop(@textonpagestack);

    open($fh,"<$thisfile") || die("can't open tempfile $thisfile");

    # suck it all in at once
    local(@thetext) = (<$fh>);

    close($fh) || die("can't close tempfile $thisfile");

    # delayed text
    if ($command eq "d") {
	$delayedtext .= "@thetext";

    # footnote
    } elsif ($command eq "f") {
	$reftext .= "@thetext";

    # index
    } elsif ($command eq "x") { 
	# this makes a 1 space named anchor. 
	# I am unsure if Mosaic can handle this or not.
	# could use invisible.xbm like texi2html
	$pretext = "<a name=\"$thisfile-mark\"> </a>";

	$currentindex = ($args) ? $args : 'x';

	# any P A arguments besides page number are thrown away
	$indextext{$currentindex} .=
	    "<dt><a href=\"$URL{'current'}#$thisfile-mark\">";
	$indextext{$currentindex} .= "@thetext</a>\n\n"; 

    #  preprocessors, like tbl, eqn, etc.
    } elsif (grep(/$command/,keys %preproc)) {
	$intable = $doubleboxhack = 0;
	if ($preproc{$command} eq "nroff") {
	    $thetext = `$command $thisfile | $nroff -me | col -b`;
	    $thetext =~ s/^\s+//g;
	    $thetext =~ s/\s+$//g;
	    $pretext =  "<pre>\n$thetext</pre>\n";
	} elsif ($preproc{$command} eq "ps") {
	    # don't bother if just delim
	    # this is a dorky hack
	    @thetext =  split("\n",join('',@thetext));
	    $lines = $#thetext + 1;
	    foreach $i (0..$#thetext) {
		$thetext = $thetext[$i];
		$lines-- if $thetext =~ /^\<\S+\>/;
		$lines-- if $thetext =~ /^\.EQ/;
		$lines-- if $thetext =~ /^\.EN/;
		$lines-- if $thetext =~ /^\s*$/;
		$lines-- if $thetext =~ /^delim/;
	    }
	    if ($lines) {
		&ps_to_gif($command, $thisfile, $args);
		# bail if this was a zero size file
		if (!-e "$thisfile.gif" || -z "$thisfile.gif") {
		    # early exit!
	    #unlink("$thisfile") || die("can't unlink tempfile $thisfile");
		#unlink("$thisfile.gif") || die("can't unlink $thisfile.gif");
		    $delid{$command}--;
		return;
		}
	    } else {
		$delid{$command}--;
		return;
	    }
	    # signal an inline eqn by $args = 'inline'
	    if ($args ne 'inline') {
		$pretext .= "<p><img src=\"$thisfile.gif\"><p>\n";
		$pretextprints = 1;
	    } else {
		# early exit!!
		unlink($thisfile) || die("can't unlink tempfile $thisfile");
		return("<img src=\"$thisfile.gif\">");
	    }
	} elsif ($preproc{$command} eq "omit") {
	    # do nothing; throw away @thetext
	}
    } else {
	warn("popdelayed error: unknown type \"$command\"\n");
    }
    unlink($thisfile) || die("can't unlink tempfile $thisfile");
}


# from latex2html pstogif script by Nikos Drakos, nikos@cbl.leeds.ac.uk
# modified version of pstoepsi script by Doug Crabill dgc@cs.purdue.edu
# pstoppm.ps from ghostscript

$pstoppm = "$troff2htmldir/pstoppm.ps";

sub ps_to_gif {
    local($command, $thisfile, $args) = @_;

    print STDERR "making postscript to gif $thisfile...\n";
    system("$preprocname{$command} $thisfile | $psroff > $thisfile.ps");
    
    system("$gs -q -dNODISPLAY $pstoppm << EOF
100 100 ppmsetdensity
($thisfile) ppm1run
EOF
");
    if ($args eq 'inline') {
      system("$pnmcrop $thisfile.ppm | $ppmtogif -trans 1,1,1 > $thisfile.gif");
    } else {
      system("$pnmcrop -right $thisfile.ppm | $pnmcrop -top | \
$pnmcrop -bottom | $ppmtogif -trans 1,1,1 > $thisfile.gif");
    }

    unlink("$thisfile.ppm", "$thisfile.ps");
}

sub mark_in_note {
    local($type) = @_;
    if ($type eq "d") { 
	$num = $currentdelay;
    } else {
	$num = $currentfoot;
    }
    return "<a name=\"$type-$num\" href=\"$URL{'current'}#$type-$num-mark\">" .
	"\\*[$num\\*]</a>";
}

# we make this a superscript foot/endnote number as a link anchor.
# it then gets processed to [#] till HTML+. People might want to go
# back and by hand change make the link a whole word or phrase instead 
# of just a stupid [#].

sub mark_outof_note {
    local($type) = @_;

    if ($type eq "d") {
	$num = ++$currentdelay;
    } else {
	$num = $currentfoot;
    }
    
    return "<a name=\"$type-$num-mark\" href=\"$URL{'notes'}#$type-$num\">" .
	"\\*[$num\\*]</a>";
}


# prints and truncates delayed text
sub printdelay {
    $pretext = "<p>$delayedtext\n"; 
    $delayedtext = "";		
    $pretextprints = 1;
}

sub printindex {
    local($args) = "x" if !$args;
    $pretext = "<dl>$indextext{$args}</dl>\n";

    $indexprinted{$args} = 1;
    $pretextprints = 1;
}

##### fonts

# this is called for rb,bi,bx,r,i,b,u
# it can be called with the letter of font
sub font {
    local($command) = @_ if @_;

    if (!@args) {
	# with no arguments, just switch font
	$pretext = &switch_font($command);
    } else {
	# with arguments, just do first word
	$pretext = &switch_font($command) . shift(@args) .
	    &switch_font("p");
	$pretext .= shift @args while @args;
	$pretext .= " "; 
	$pretextprints = 1;
    }
}

# html eats spaces in <tt>; assume if did a .ft, want <pre> instead
sub change_font {
    local($command) = @_ if @_;

    $command = "pre" if ($command eq "C");
    $pretext = &switch_font($command);
}

# this outputs the html that will do the switch
# no 'environments' are implemented
# I am assuming fonts switch back to normal at pp. true?
# I am assuming fonts switches inside blocks are local to block. true?
# keep a @fontstack

sub switch_font {
    local($newfont) = @_ if @_;
    local($result) = "";

    # remember, .fp is ignored for now. S font not available
    # if called with a number 1-3
    $newfont = ("r","i","b")[$newfont-1] if ($newfont =~ /\d/);
    $newfont =~ tr/A-Z/a-z/;

    if (@fontstack) {
	$result .= $endcodes{pop @fontstack};
    } else {
	$result .= $endcodes{"r"};
    }

    if ($newfont eq "p") {
	if ($#fontstack >= 0) {
	    $newfont = $fontstack[$#fontstack];
	} else {
	    $newfont = "r";
	}
    }
    if (!defined($begincodes{$newfont})) { 
	die("undefined font: $newfont! at $command $args\n$text\n");
    }
    $result .= $begincodes{$newfont};
    push(@fontstack,$newfont);
    #$result .= "<br><b>newfont: $newfont; stack: @fontstack</b><br>\n";
    return $result;
}

sub pop_all_fonts {
    local($font, $result);

    while ($font = pop @fontstack) {
	#$result .= "<br><b>popped $font; stack: @fontstack</b><br>\n";
	$result .= $endcodes{$font};
    }
    return $result;
}

%begincodes = ("r","",
	       "i","<i>",
	       "b","<b>",
	       "bi","<b><i>",
	       "c","<code>",
	       "t","<tt>",
	       "pre","<pre>",
	       "h","",   # helvetica?
	       "p","");

	       
%endcodes = ("r","",
	     "i","</i>",
	     "b","</b>",
	     "bi","</i></b>",
	     "c","</code>",
	     "t","</tt>",
	     "pre","</pre>",
	     "h","",
	     "p","");

### preprocessors

# used for eqn, tbl, etc.
# should be given commmand "eqn", "tbl", not "EQ", "TS", etc.
# ignores @args
sub beginpreproc {
    local($command) = @_ if @_;

    # pop any pending lists BUT NOT FONTS (at least for eqn)
    # $pretext .= &pop_all_lists;

    if ($preproc{$command} eq "pre") {
	# just give the raw commands
	$pretext .= "<pre>\n";
    } elsif ($preproc{$command} =~ /(nroff|ps|omit)/) {
	# first check for delim commands
	if ($command eq 'eqn') {
	    if ($text =~ /delim off/) {
		$eqnopendelim = $eqnclosedelim = 0;
	    } elsif ($text =~ /delim (\S)(\S)/) {
		# can only be single char
		$eqnopendelim = $1;
		$eqnclosedelim = $2;
		$eqnopendelim =~ s/(\W)/\\\1/;
		$eqnclosedelim =~ s/(\W)/\\\1/;
	    }
	}
	# push the command we just read to text, not pretext...
	$text = $preprocbegin{$command} . $text;
	    
	&pushdelayed($command);
    } else {
	die("unsupported $command option $preproc{$command}\n");
    }
}

# put back on the command that we changed in the first place...
# I'm not sure why I'm doing this backflip...
%preprocend = ("eqn", ".EN",
	       "tbl", ".TE",
	       "pic", ".PE");

# this immediately breaks -ms EQ...
%preprocbegin = ("eqn", "\n.EQ\n",
		 "tbl", "\n.TS\n",
		 "pic", "\n.PS\n");
sub endpreproc {
    local($command) = @_ if @_;

    if ($preproc{$command} eq "pre") {
	$pretext = "</pre>\n";
    } elsif ($preproc{$command} =~ /(nroff|ps|omit)/) {
	# shell processing is going to need the end flag
	&output($preprocend{$command},1);
	&popdelayed($command);
    } else {
	die("unsupported $command option $preproc{$command}\n");
    }
}


### misc

sub horizline {
    $pretext = "<hr>";
}

###### raw troff

### font

# specific to .cs syntax
sub constantwidth {
    local(@args) = @_ if @_;

    shift @args; # get rid of F (font)

    # if anything (N) left, turn on, else turn off
    &switch_font( @args ? "t" : "p");
}

### text filling

sub breakline {
    $pretext = "<br>\n";
}

# we either break lines explicitly or switch into <pre> mode
# call this with 1 to break, 0 to not break (fill)
# nf/fi are serious changes, and last till undone; thus the $nofill variable
sub fillmode {
    local($mode) = @_;

    if ($opt_nofill eq "br") {
	$breaklines = $mode;
    } else {
	$pretext = $mode ? "<pre>" : "</pre>";
	$nofill = $mode;
    }
}

# can't center yet, so do <dl><dd> instead
# resort to horrible bookkeeping to count physical lines
sub center {
    $pretext = &pop_all_lists;
    $args = 1 if ($args eq '');
    if ($args != 0) {
	# begin center
	$pretext .= "<dl><dd>";
	$indentlines = $args;
	push(@breakstack,$breaklines);
	$breaklines = 1;
    } else {
	# end center
	$indentlines = 0;
	$breaklines = pop @breakstack;
	$pretext .= "</dl>";
    }
}


### line length

# indent macros goe here. Use <dl><dd>, pop_list to change.
# only do it if $textonpage, otherwise assume somebody's just
# fiddling with the margins :-)

### macros

# macros are not supported, although I think they could be

# macros can be nested as long as enddelim is different
sub macro {
    local(@args) = @_ if @_;
    local($macro) = shift @args;
    local($enddelim);
    if (@args) {
	$enddelim = $args[0];
    } else {
	$enddelim = '..';
    }

    # delete leading dot?
    $enddelim =~ s/^\.//;

    # escape nonword chars, since this will be used in a regexp
    $enddelim =~ s/(\W)/\\\1/g;

    warn("ignoring definition for $macro\n");

    push(@macro, $enddelim);
}
    
# strings are supported
sub string {
    local(@args) = @_ if @_;

    local($key) = shift @args;
    local($val) = join(" ",@args);

    # escape all the non-word chars
    $key =~ s/(\W)/\\\1/g;
    # can't just go in and escape val, since size mods are already'
    # escaped \\s-1 etc.
    $val =~ s/(\W)/\\\1/g;
    $val =~ s/\\\\\\\\/\\\\/g; # if \\\\, \\

    # note we always append
    $strings{$key} .= $val;

    &make_stringeval;
}

sub remove {
    local(@args) = @_ if @_;

    delete $strings{$_};
    &make_stringeval;
}

sub rename {
    @names = split;
    $strings{$names[1]} = $strings{$names[0]};
    delete $strings{$names[0]};
    &make_stringeval;
}


sub make_stringeval {
    # now we need a form that's an eval-able perl string '
    $stringeval = "";
    while (($key,$val) = each %strings) {
	if (length($key) == 1) {
	    $stringeval .= "s/\\\\\\\*$key/$val/g;\n" if ($key ne "");
	} else {
	    $stringeval .= "s/\\\\\\\*\\\($key/$val/g;\n" if ($key ne "");
	}
    }				
}


##### i/o file switching

# full recursive file sourcing
sub source {
    $source++;

    # local filehandle
    local($sourcefile) = "SOURCE$source";
    open($sourcefile,$args) || 
	warn("can't open .so source file: $!\n");
    warn("sourcing file $args\n");

    # need a local copy of the rest of our current record
    local($oldtext) = $text;
    while(<$sourcefile>) {
	&handle_record;
    }
    close($sourcefile);
    $text = $oldtext;
}

sub ignore {
}

sub echocommand {
    $pretext .= $command;
    $pretextprints = 1;
}

1;
